/*
** Module   : DmoWriteManager.java
** Abstract : Stores and manages DmoWriteStorages. 
**
** Copyright (c)2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------- Description --------------------------------
** 001 OM  20140425 Created initial version with minimum required functionality.
*/

package com.goldencode.p2j.persist;

import java.util.*;

/**
 * This class does the management for DmoWriteStorages objects. They can be looked up, created
 * and removed when the dmo and persistences are not in use.
 * To accomplish that the class uses a double hashing map using as keys the persistence objects 
 * and dmos.
 * 
 * A single instance of this class is created for each connection and can be obtained from 
 * <code>BufferManager</code> instance. 
 */
public class DmoWriteManager
{
   Map<Persistence, Map<Persistable, DmoWriteStorage>> data = new HashMap<>();

   /**
    * Obtain the DmoWriteStorage that handles the writes and undo events for specified 
    * persistence and DMO. If such object is not found it can be optionally created.
    * 
    * @param   p
    *          The persistence. 
    * @param   dmo
    *          The dmo to be persisted.
    * @param   create
    *          Specifies if the handle should be created if not found in the internal structure.
    *          
    * @return  The requested DmoWriteStorage or <code>null</code> if not found and creation is 
    *          not required.  
    */
   DmoWriteStorage getStorage(Persistence p, Persistable dmo, boolean create)
   {
      Map<Persistable, DmoWriteStorage> map = data.get(p);
      if (map == null)
      {
         if (!create) 
         {
            // creation is not requested so we can return null now
            return null;
         }

         // must create it now
         map = new HashMap<>();
         data.put(p, map);
      }
      assert (map != null); // evidently
      
      DmoWriteStorage storage = map.get(dmo);
      if (storage == null)
      {
         if (!create)
         {
            // creation is not requested so we can return null now
            return null;
         }
         storage = new DmoWriteStorage(dmo.getClass().getSimpleName());
         map.put(dmo, storage);
      }
      
      return storage;
   }
   
   /**
    * It is perfectly normal to call remove on a dmo that wasn't previously stored if the 
    * respective dmo is the content for a temporary, read-only buffer used in WRITE triggers. 
    */
   void remove(Persistence p, Persistable dmo)
   {
      Map<Persistable, DmoWriteStorage> map = data.get(p);
      if (map == null)
      {
         return;
      }

      map.remove(dmo);
      
      // little cleanup
      if (map.isEmpty())
      {
         data.remove(p);
      }
   }

   /**
    * Removes persistence <code>p</code> from internal data structure. All information about DMOs
    * that depends on it are also removed. There is no listener for this action and the removed
    * information is lost.
    * It does not issues any error is such object is not found in the internal data structure.
    * 
    * @param   p
    *          The Persistence to be removed. 
    */
   void remove(Persistence p)
   {
      data.remove(p);
   }

   /**
    * Notifies all known <code>DmoWriteStorage</code>s that a block is left.  
    * 
    * @param   blockDepth
    *          The depth of the block left.
    */
   void exitBlock(int blockDepth)
   {
      for (Persistence p : data.keySet())
      {
         Map<Persistable, DmoWriteStorage> map = data.get(p);
         for (Persistable dmo : map.keySet())
         {
            DmoWriteStorage storage = map.get(dmo);
            storage.exitBlock(blockDepth);
         }
      }
   }

   /**
    * Notifies all known <code>DmoWriteStorage</code>s the start of a new block.   
    *
    * @param   blockDepth
    *          The depth of the block that starts.
    */
   void enterBlock(int blockDepth)
   {
      for (Persistence p : data.keySet())
      {
         Map<Persistable, DmoWriteStorage> map = data.get(p);
         for (Persistable dmo : map.keySet())
         {
            DmoWriteStorage storage = map.get(dmo);
            storage.enterBlock(blockDepth);
         }
      }
   }
}
