package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.ErrorManager;
import java.util.LinkedList;

public class DmoWriteStorage
{
   /**
    * The last persisted record to database. This is needed for WRITE trigger to be optionally 
    * passed as old buffer value and checking if the content of the current record
    * is changed when rollbacking. It may be null if the buffer is not initialized or persisted.
    */
   private Persistable lastPersistedDMO = null;

   /** Flags that this buffer was changed since last write to database. */
   private boolean wasTouched = false;
   // TODO: investigate if (wasTouched) can be replaced by (lastPersistedDMO != null)

   /**
    * This stack contains the data needed for flush method in order to decide if the WRITE
    * database events / triggers should be fired and management in the case of an UNDO statement.
    * The lastPersistedValue and wasTouched evolve independently of the current executing block
    * but in the event of UNDO they are restored from this stack.
    */
   private LinkedList<WriteSnapshot> flushSnapshotData = new LinkedList<>();

   private String recordName;

   public DmoWriteStorage(String recordName)
   {
      this.recordName = recordName;
      logMe("created", "");
   }

   public boolean rollback(Persistence persistence, int blockDepth)
   {
      logMe("rollback", blockDepth + "");
      
      if (flushSnapshotData.isEmpty())
      {
         logMe("removing ? ", "no such block " + blockDepth);
      }

      boolean forceFlush = false; 
      final WriteSnapshot flushNode = flushSnapshotData.getLast(); // the state before entering this block
      assert (flushNode.blockDepth == blockDepth); // sanity check

      // NOTE: we are intentionally using the == operator. There is no need to check their
      // content bit-by-bit because the DMO snapshot is built from the scratch each time
      // the flush method decides to save the DMO to persistence
      if (lastPersistedDMO == flushNode.persistedDMO)
      {
         // no commits of the record in this block ? restore the wasTouched from upper block
         wasTouched = flushNode.dirty;
      }
      else
      {
         // there was at least one flush/commit to database, restoring the old content:

         // rolling back to value at the beginning of the block
         lastPersistedDMO = flushNode.persistedDMO;
         wasTouched = flushNode.dirty;

         // if lastPersistedDMO == null then the record was newly created,
         // nothing to restore to database, the ReversibleCreate will handle if the case
         if (lastPersistedDMO != null)
         {
            boolean commit = false;
            try
            {
               if (false) // disable silent db rollbacking
               {
                  // Persist back the old snapshot.
                  commit = persistence.beginTransaction();

                  // TODO: using Session.save() will result in a "duplicate key value violates unique constraint"
                  // TODO: using Session.saveOrUpdate() assignemnt will fail.
                  persistence.save(lastPersistedDMO, lastPersistedDMO.getId());
               }

               // validation is not be necessary as the currently saved record was persisted
               // successfully some time ago
               forceFlush = !true;
            }
            catch (PersistenceException exc)
            {
               if (commit)
               {
                  commit = false;
                  try
                  {
                     persistence.rollback();
                  }
                  catch (PersistenceException exc2)
                  {
                  }
               }
               ErrorManager.recordOrThrowError(exc);
            }
            finally
            {
               if (commit)
               {
                  try
                  {
                     persistence.commit();
                  }
                  catch (PersistenceException exc)
                  {
                     ErrorManager.recordOrThrowError(exc);
                  }
               }
            }
         }
      }
      return forceFlush;
   }

   public void touch(Persistable currentRecord)
   {
      logMe("touch", currentRecord.toString());
      wasTouched = true;

      if (lastPersistedDMO == null)
      {
         // very lazy initialization, this is the first time the record was changed
         lastPersistedDMO = (Persistable) currentRecord.deepCopy();
      }
   }

   public Persistable getLastPersistedDMO()
   {
      return lastPersistedDMO;
   }

   public boolean isWasTouched()
   {
      return wasTouched;
   }

   public void setLastPersistedDMO(Persistable lastPersistedDMO)
   {
      logMe("setLastPersistedDMO", lastPersistedDMO.toString());
      this.lastPersistedDMO = lastPersistedDMO;
   }

   public void setWasTouched(boolean wasTouched)
   {
      logMe("setWasTouched", wasTouched + "");

      this.wasTouched = wasTouched;
   }

   public boolean needFlush()
   {
      return (lastPersistedDMO != null) && wasTouched;
   }

   public void resetFlushData()
   {
      wasTouched = false;
      lastPersistedDMO = null;
   }

   public void enterBlock(int blockDepth)
   {
      logMe("enterBlock", blockDepth + "");
      flushSnapshotData.addLast(new WriteSnapshot(blockDepth, wasTouched, lastPersistedDMO));
   }

   public void exitBlock(int blockDepth)
   {
      logMe("exitBlock", blockDepth + "");
      if (flushSnapshotData.isEmpty())
      {
         // unregister me, probably out of scope!
         logMe("removing ? ", "no such block " + blockDepth);
         
         return;
      }
      final WriteSnapshot last = flushSnapshotData.removeLast();
      assert (last.blockDepth == blockDepth);
   }

   private void logMe(String method, String info1)
   {
      System.err.println("DMOStorage::" + method + "(" + recordName + ", " + info1 + "): " +
             (lastPersistedDMO == null ? "null" : lastPersistedDMO.getId().toString()));
   }
   
   
   /**
    * This immutable class is an element in the stack of undo nodes for previous persisted DMOs.
    * The objects of this kind are used only in a rollback operation in order to obtain the exact
    * configuration of the OLD BUFFER in a WRITE trigger. Because WRITE event is not directly
    * related to transactions block we cannot rely on the snapshot field of RecordBuffer.
    */
   class WriteSnapshot
   {
      /**
       * The block depth (BD) at which the snapshot is taken. This field of the class is just
       * for debugging purposes.
       */
      private int blockDepth;

      /**
       *  Flag for checking if the record touched before entering block (in BD-1).
       */
      private boolean dirty;

      /**
       * The value of lastPersistedValue entering block (in BD-1).
       */
      private Persistable persistedDMO;

      /**
       * The constructor that builds the object.
       *
       * @param   blockDepth
       *          The block depth (BD) at which the snapshot is taken.
       * @param   dirty
       *          Flags if the record touched before entering block.
       * @param   persistedDMO
       *          The value of lastPersistedValue entering block.
       */
      WriteSnapshot(int blockDepth, boolean dirty, Persistable persistedDMO)
      {
         this.blockDepth = blockDepth;
         this.dirty = dirty;
         this.persistedDMO = persistedDMO;
      }
   }
}


